//  Software License Agreement
//
// The software supplied herewith by Microchip Technology Incorporated (the "Company")
// for its PICmicro Microcontroller is intended and supplied to you, the Company?s
// customer, for use solely and exclusively on Microchip PICmicro Microcontroller
// products.
//
// The software is owned by the Company and/or its supplier, and is protected under
// applicable copyright laws. All rights are reserved. Any use in violation of the
// foregoing restrictions may subject the user to criminal sanctions under applicable
// laws, as well as to civil liability for the breach of the terms and conditions of
// this license.
//
// THIS SOFTWARE IS PROVIDED IN AN "AS IS" CONDITION. NO WARRANTIES, WHETHER EXPRESS,
// IMPLIED OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE. THE
// COMPANY SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL OR
// CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.

/**********************************************************************************
 * This very simple program shows how to configure the PWMs on the PIC16F1503 and
 * use them to control a High Brightness (HB) RGB LED. When using an RGB LED you will
 * most likely need to use some kind of lens or other diffuser to allow the different
 * colors to "mix" and produce a single resulting color.
//*****************************************************************************************

*******************************************************************************************
*  Device    : PIC16F1503
*  Compiler  : XC8 v1.30
*  MPLAB     : MPLAB X 2.20
*
***  Note: The PIC16F1503 does NOT allow in-circuit debugging - only in-circuit programming
********************************************************************************************
*/


// This program controls 3 HB LEDs using 3 of the PWM modules.
// *****************************************************************************************
#include <xc.h> // include standard header file


// set Config bits

// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection Bits (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config BOREN = ON       // Brown-out Reset Enable (Brown-out Reset enabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LPBOR = OFF      // Low-Power Brown Out Reset (Low-Power BOR is disabled)
#pragma config LVP = ON         // Low-Voltage Programming Enable (Low-voltage programming enabled)



// Definitions
#define _XTAL_FREQ  16000000            // this is used by the __delay_ms(xx) and __delay_us(xx) functions

#define led_Red    LATCbits.LATC5	// Red LED
#define led_Green  LATCbits.LATC3	// Green LED
#define led_Blue   LATAbits.LATA2	// Blue LED

#define pwm_red 1       // these are used to simplify calls to the Set_PWM_Duty routine
#define pwm_green 2
#define pwm_blue 3


// global arrays for the linear interpolation table.
// These table values were created based on the fact that the brightness of LEDs is typically not
// linear with respect to PWM duty cycle and most LEDs require a certain minimum duty cycle
// before they generate any light.  These values may need to be changed for your LEDs.
// The values in the table below were generated from the equation
// y = 9E-07x3 - 9E-05x2 + 0.105x + 45 where x is the incoming PWM value between 0 and 1023 and
// y is the compensated PWM value.
const unsigned int x_interp[14] = {0,50,100,150,200,250,300,400,500,600,750,800,900,1023};
const unsigned int y_interp[14] = {45,50,56,62,70,80,93,130,188,270,453,532,723,1022};
const unsigned char max_array = 13; // array elements are 0 to 13

//**************************************************************************************
// delay x number of millseconds
//**************************************************************************************
void mydelay_ms(unsigned int cycles) {
    unsigned int i;

    for (i = 1; i <= cycles; i++) {
        __delay_ms(1);
    }
}
//**************************************************************************************
// delay x number of seconds
//**************************************************************************************
void mydelay_sec(unsigned int cycles) {
    unsigned char i;

    for (i = 1; i <= cycles; i++) {
        mydelay_ms(1000);
    }
}

//**************************************************************************************
// Linearize LED Brightness (beta)
// Use linear interpretation method to linearize the curve for pwm value vs. LED brightness.
// For more information you can do an internet search for "LED Brightness Curve"
//**************************************************************************************
unsigned int LinearizedPWM(unsigned int PWMinput)
{
    unsigned int x,x0,x1,y0,y1;   //known points on curve (from table)
    unsigned int compensated_pwm; // calculated pwm to compensate for curve
    unsigned char i;
    unsigned char valueFound;


    // first deal with input pwm values of zero or less
    if (PWMinput <= 0)
        compensated_pwm = y_interp[0];  // grab first value in table and jump out
    else
    {
        // first find the x value (pwm input) point in the array that matches or is smaller
        // then the given input
        i=0;
        valueFound = 0;

        while (!valueFound)
        {
            if (x_interp[i+1] >= PWMinput ) // find first value in table equal or smaller than input pwm
            {
                 valueFound = 1;
                 if (x_interp[i+1] == PWMinput) // if they are equal then we already know the y value
                     compensated_pwm = y_interp[i+1];
                 else // we have to calculate
                 {
                    x = PWMinput;       
                    x0 = x_interp[i];   
                    y0 = y_interp[i];   
                    x1 = x_interp[i+1]; 
                    y1 = y_interp[i+1]; 
                    compensated_pwm = y0 + ((y1-y0)*(x - x0))/(x1 - x0);
                   
                 }
            }
            else
            {
                if ((i+1)>= max_array)// then we use last values in the table
                {
                    compensated_pwm = y_interp[max_array];
                    valueFound = 1;
                }
                else
                    i++; // jump to next set of values in the array
            }

        }   // end of while loop
    } // end of if loop

    
    return (compensated_pwm);
}



//*************************************************************************************
// Set PWM Duty
// Set LED channel and PWM duty cycle
// Inputs:
// PWM Channel = value 1 (red), 2 (green) or 3 (blue)
// pwm duty = 10 bit duty cycle value between 0 and 1023
//**************************************************************************************
void Set_PWM_Duty(unsigned char PWM_Channel,unsigned int pwm_duty)
{
    
    // NOTE: if you don't want to use the linearization routine, comment out this next line
    pwm_duty = LinearizedPWM(pwm_duty); 

    if (PWM_Channel == 1)
    {
        PWM1DCH = pwm_duty >> 2;        // take input duty cycle value and divide into high bits
        PWM1DCL = pwm_duty & 0x0003;    // and low bits of the duty cycle registers
    }
    else if (PWM_Channel == 2)
    {
        PWM2DCH = pwm_duty >> 2;
        PWM2DCL = pwm_duty & 0x0003;
    }
    else if (PWM_Channel == 3)
    {
        PWM3DCH = pwm_duty >> 2;
        PWM3DCL = pwm_duty & 0x0003;
    }
}



//**************************************************************************************
// Ramp LED
// This routine ramps brightness of one LED up or down by changing pwm duty cycle.
//
//**************************************************************************************
// inputs:
// unsigned char channel - channel 1(red) or 2(green) or 3(blue)
// int16 startval - starting duty cycle value
// int16 endval - ending duty cycle value (0 - 1023) NOTE: make sure you step size divides evenly into this value.
// int16 stepsize - size of duty cycle step while ramping (1 - whatever...use values of 1 for very slow and 20 or 50 for fast ramps)
// int16 delaycount -  delay in ms between each step (I usually use values of 5 or 10)
//**************************************************************************************
void RampLED(unsigned char channel,unsigned int startval,unsigned int endval,unsigned int stepsize, unsigned int delaycount)
{
    unsigned int count;

    if (endval > startval)
    {
        for (count=startval;count<endval;count=count+stepsize)
        {
            Set_PWM_Duty(channel,count);
            mydelay_ms(delaycount);
        }
    }
    else
    {
        for (count=startval;count>endval;count=count-stepsize)
        {
            Set_PWM_Duty(channel,count);
            mydelay_ms(delaycount);
        }
    }
}


//*************************************************************************************
void main() {
//*************************************************************************************
    unsigned int pwm_red_count;
    unsigned int pwm_green_count;
    unsigned int pwm_blue_count;
    unsigned int i,step;

    
   
    // set up oscillator control register
    OSCCONbits.IRCF = 0x0F; //set OSCCON IRCF bits to select OSC frequency=16Mhz
    OSCCONbits.SCS = 0x02;  //set the SCS bits to select internal oscillator block
    LATA=0;                 // turn all outputs off
    LATC=0;

    
    
    OPTION_REGbits.nWPUEN = 1; // all weak pullups are disabled

    // PORT A Assignments
    TRISAbits.TRISA0 = 1; // RA0 = nc (SPDAT)
    TRISAbits.TRISA1 = 1; // RA0 = nc (SPCLK)
    TRISAbits.TRISA2 = 1; // RA2 = Blue LED
    //TRISAbits.TRISA3 = 1; // RA0 = nc (MLCR)
    TRISAbits.TRISA4 = 0; // RA4 = nc
    TRISAbits.TRISA5 = 0; // RA5 = nc

    ANSELA = 0x00; // all port A pins are digital I/O

    // PORT C Assignments
    TRISCbits.TRISC0 = 0; // RC0 = nc
    TRISCbits.TRISC1 = 0; // RC1 = nc
    TRISCbits.TRISC2 = 0; // RC2 = nc
    TRISCbits.TRISC3 = 1; // RC3 = Green LED
    TRISCbits.TRISC4 = 1; // RC4 = Mode button input
    TRISCbits.TRISC5 = 1; // RC5 = Red LED
    
    ANSELC = 0x00; // all port C pins are digital I/O


    // Setup PWMs in specific order as per datasheet
    // ***************************************************
    PWM1CON=0;     // clear PWM1 control register
    PWM2CON=0;     // clear PWM2 control register
    PWM3CON=0;     // clear PWM3 control register
    
    PWM1CONbits.PWM1POL = 1;     // polarity bit set high so Duty cyle of 1023 is full on
    PWM2CONbits.PWM2POL = 1;     // polarity bit set high so Duty cyle of 1023 is full on
    PWM3CONbits.PWM3POL = 1;     // polarity bit set high so Duty cyle of 1023 is full on

    PR2 = 0xFF;

    PWM1DCH = 0x00;             // start PWM 1 duty cycle at 0
    PWM1DCL = 0x00;
    PWM2DCH = 0x00;             // start PWM 2 duty cycle at 0
    PWM2DCL = 0x00;
    PWM3DCH = 0x00;             // start PWM 3 duty cycle at 0
    PWM3DCL = 0x00;

    PIR1bits.TMR2IF=0;		// clear TMR2 interrupt flag
    T2CONbits.T2CKPS = 0x02;    // TMR2 prescalar bits are both high; prescalar = 64

    PWM1CONbits.PWM1EN = 1;     // enable PWM1 module
    PWM2CONbits.PWM2EN = 1;     // enable PWM2 module
    PWM3CONbits.PWM3EN = 1;     // enable PWM3 module

    T2CONbits.TMR2ON=1;		// TMR2 on

    // Set LED tris bits to output
    TRISAbits.TRISA2 = 0;       // RA2 = Blue LED
    TRISCbits.TRISC3 = 0;       // RC3 = Green LED
    TRISCbits.TRISC5 = 0;       // RC5 = Red LED
    
    //...and turn on PWM output enable pins
    PWM1CONbits.PWM1OE = 1;     // enable the output for PWM1
    PWM2CONbits.PWM2OE = 1;     // enable the output for PWM2
    PWM3CONbits.PWM3OE = 1;     // enable the output for PWM3

    //***********************************************************************

    Set_PWM_Duty(pwm_red,0);   // start with all LEDs off
    Set_PWM_Duty(pwm_green,0);
    Set_PWM_Duty(pwm_blue,0);
    
    //*********************************************************************
    // NOTE!!: startval and endval numbers must be a divisible by stepsize
    // NOTE: raw pwm values below 60 will usually not turn LEDs on
    //*********************************************************************
    while(1)
    {

        // blink each color once
        Set_PWM_Duty(pwm_red,1023); // set red at full brightness
        mydelay_ms(500);            // delay .5 sec
        Set_PWM_Duty(pwm_red,0);    // set red off
        mydelay_ms(500);

        Set_PWM_Duty(pwm_green,1023);  // set green at full brightness
        mydelay_ms(500);               // delay .5 sec
        Set_PWM_Duty(pwm_green,0);     // set green off
        mydelay_ms(500);

        Set_PWM_Duty(pwm_blue,1023);   // set blue at full brightness
        mydelay_ms(500);               // delay .5 sec
        Set_PWM_Duty(pwm_blue,0);      // set blue off
        mydelay_ms(500);


        //**************************************************************
        // ramp individual colors up and down
        // For each color, use a step size of 5 and delay of 10ms
        // between each step
        //**************************************************************
        
        RampLED(pwm_red,0,1020,5,10);	// ramp red up from 0 to 1020 in steps of 5 and delay of 10ms between each step
        mydelay_ms(500);
        RampLED(pwm_red,1020,0,5,10);	// ramp red down
        mydelay_ms(500);
       
        RampLED(pwm_green,0,1020,5,10);	// ramp green up
        mydelay_ms(500);
        RampLED(pwm_green,1020,0,5,10);	// ramp green down 
        mydelay_ms(500);

        RampLED(pwm_blue,0,1020,5,10);	// ramp blue up
        mydelay_ms(500);
        RampLED(pwm_blue,1020,50,5,10);	// ramp blue down
        mydelay_ms(500);

        
        // To ramp more than one color at a time you can't use
        // the RAMPLED routine - you have to create your own loops as
        // shown below using the Set_PWM_Duty routine.
        //
        // NOTE!!: startval and endval numbers must be divisible by stepsize or
        // you will end up out in the weeds
      
        
        //******************
        //start this section by ramping up red to max
        RampLED(pwm_red,0,1020,5,5);	// ramp red up to max

        //now set up to ramp down red and ramp up green at the same time
        
        pwm_red_count = 1020;                       // set starting duty cycle for red
        pwm_green_count=0;                          // set starting duty cycle for green
        step=2;                                     // set step size
        for (i=0;i<=1020;i=i+step)
        {
            Set_PWM_Duty(pwm_red,pwm_red_count-i);  // ramp red down
            Set_PWM_Duty(pwm_green,i);              // and green up at the same time
            mydelay_ms(10);                         // delay
        }
        
        pwm_red_count = 0;                          // make sure red is off 
        pwm_green_count=1020;                       // set starting duty cycle for green
        pwm_blue_count=0;                           // set starting duty cycle for blue
        step=2;
        for (i=0;i<=1020;i=i+step) 
        {
            Set_PWM_Duty(pwm_green,pwm_green_count-i);  // ramp green down and blue up at same time
            Set_PWM_Duty(pwm_blue,i);
            mydelay_ms(10);
        }

        pwm_red_count = 0;
        pwm_green_count=0;
        pwm_blue_count=1020;
        step=2;
        for (i=0;i<=1020;i=i+step) 
        {
            Set_PWM_Duty(pwm_blue,pwm_blue_count-i);    // ramp blue down and red up at same time
            Set_PWM_Duty(pwm_red,i);
            mydelay_ms(10);
        }

        mydelay_ms(500);
       
    

    } // end of while loop
    

   

} // end main routine











